/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// Main windowed and fullscreen graphics interface module. This module
// is used for both the software and OpenGL rendering versions of the
// Quake refresh engine.
#include <assert.h>
#include <float.h>

#include "client.h"
#include "winquake.h"
#include "d3dw_win.h"

int MapKey (int key);


// Global variables used internally by this module

D3DDISPLAYMODE	viddef;				// global video state; used by other modules
HWND        cl_hwnd;            // Main window handle for life of program

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

extern	unsigned	sys_msg_time;


/*
==========================================================================

DLL GLUE

==========================================================================
*/

#define	MAXPRINTMSG	4096

void VID_Printf (int print_level, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
	static qboolean	inupdate;

	va_start (argptr, fmt);
	vsprintf (msg, fmt, argptr);
	va_end (argptr);

	if (print_level == PRINT_ALL)
	{
		Com_Printf ("%s", msg);
	}
	else if (print_level == PRINT_DEVELOPER)
	{
		Com_DPrintf ("%s", msg);
	}
	else if (print_level == PRINT_ALERT)
	{
		MessageBox (0, msg, "PRINT_ALERT", MB_ICONWARNING);
		OutputDebugString (msg);
	}
}

void VID_Error (int err_level, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
	static qboolean	inupdate;

	va_start (argptr, fmt);
	vsprintf (msg, fmt, argptr);
	va_end (argptr);

	Com_Error (err_level, "%s", msg);
}


void AppActivate (BOOL fActive, BOOL minimize)
{
	Minimized = minimize;

	Key_ClearStates ();

	// we don't want to act like we're active if we're minimized
	if (fActive && !Minimized)
		ActiveApp = true;
	else ActiveApp = false;

	// minimize/restore mouse-capture on demand
	if (!ActiveApp)
	{
		IN_Activate (false);
		CDAudio_Activate (false);
		S_Activate (false);
	}
	else
	{
		IN_Activate (true);
		CDAudio_Activate (true);
		S_Activate (true);
	}
}

/*
====================
MainWndProc

main window procedure
====================
*/
LONG CDAudio_MessageHandler (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void IN_ReadRawInput (HRAWINPUT hRawInput);
void IN_UpdateClipRect (void);
void VID_HandleAltEnter (void);

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INPUT:
		IN_ReadRawInput ((HRAWINPUT) lParam);
		break;

	case WM_ERASEBKGND: return 1; // treachery!!! see your MSDN!

	case WM_MOUSEWHEEL:
		// this chunk of code theoretically only works under NT4 and Win98
		// since this message doesn't exist under Win95
		if ((short) HIWORD (wParam) > 0)
		{
			Key_Event (K_MWHEELUP, true, sys_msg_time);
			Key_Event (K_MWHEELUP, false, sys_msg_time);
		}
		else
		{
			Key_Event (K_MWHEELDOWN, true, sys_msg_time);
			Key_Event (K_MWHEELDOWN, false, sys_msg_time);
		}

		return 0;

	case WM_HOTKEY:
		return 0;

	case WM_CREATE:
		cl_hwnd = hWnd;
	case WM_MOVE:
	case WM_SIZE:
		IN_UpdateClipRect ();
		break;

	case WM_DESTROY:
		// let sound and input know about this?
		cl_hwnd = NULL;
		break;

	case WM_ACTIVATE:
		{
			int	fActive, fMinimized;

			// KJB: Watch this for problems in fullscreen modes with Alt-tabbing.
			fActive = LOWORD (wParam);
			fMinimized = (BOOL) HIWORD (wParam);

			AppActivate (fActive != WA_INACTIVE, fMinimized);
			GLimp_AppActivate (!(fActive == WA_INACTIVE));

			return 0;
		}

	// this is complicated because Win32 seems to pack multiple mouse events into
	// one update sometimes, so we always check all states and look for events
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
	case WM_MBUTTONDOWN:
	case WM_MBUTTONUP:
	case WM_MOUSEMOVE:
		{
			int	temp = 0;

			if (wParam & MK_LBUTTON) temp |= 1;
			if (wParam & MK_RBUTTON) temp |= 2;
			if (wParam & MK_MBUTTON) temp |= 4;

			IN_MouseEvent (temp);

			return 0;
		}

	case WM_SYSCOMMAND:
		if (wParam == SC_SCREENSAVE)
			return 0;

		break;

	case WM_SYSKEYDOWN:
		if ((int) wParam == 13)
		{
			VID_HandleAltEnter ();
			return 0;
		}

		// fall through
	case WM_KEYDOWN:
		Key_Event (MapKey (lParam), true, sys_msg_time);
		return 0;

	case WM_SYSKEYUP:
	case WM_KEYUP:
		Key_Event (MapKey (lParam), false, sys_msg_time);
		return 0;

	case MM_MCINOTIFY:
		CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
		return 0;

	default:
		// pass all unhandled messages to DefWindowProc
		break;
	}

	// return 0 if handled message, 1 if not
	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}


/*
** VID_UpdateWindowPosAndSize
*/
void VID_UpdateWindowPosAndSize (int x, int y)
{
}


/*
** VID_NewWindow
*/
void VID_NewWindow (int width, int height)
{
	viddef.Width = width;
	viddef.Height = height;

	cl.force_refdef = true;		// can't use a paused refdef
}


/*
============
VID_Init
============
*/
void VID_Init (void)
{
	// Add some console commands that we want to handle
	Cmd_AddCommand ("vid_restart", VID_ResetDevice);

	// Start the graphics mode and load refresh DLL
	cl.refresh_prepped = false;
	cls.disable_screen = true;

	// ensure that we have the modelist up
	D3D_InitModeList ();

	// init endianness buggery (not sure why this is here as it's already in Qcommon_Init too)
	Swap_Init ();

	if (!R_Init (GetModuleHandle (NULL), MainWndProc))
	{
		R_Shutdown ();
		Com_Error (ERR_FATAL, "Couldn't load D3D refresh!");
		return;
	}

	Com_Printf ("------------------------------------\n");
	cls.disable_screen = false;
}


/*
============
VID_Shutdown
============
*/
void VID_Shutdown (void)
{
	R_Shutdown ();
}


